/*
 * crash-manager
 *
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <dirent.h>
#include <fcntl.h>
#include <getopt.h>
#include <gio/gio.h>
#include <iniparser.h>
#include <inttypes.h>
#include <libgen.h>
#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/procfs.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <sys/vfs.h>
#include <unistd.h>

#include <pkgmgr-info.h>
#include <tzplatform_config.h>

#undef LOG_TAG
#define LOG_TAG "CRASH_MANAGER"

#include "defs.h"
#include "shared/log.h"
#include "shared/config.h"
#include "shared/spawn.h"
#include "shared/util.h"
#include "so-info.h"
#include "crash-manager.h"

/* Parsing */
#define KEY_MAX              255

#define CRASH_TEMP_SUBDIR    "/temp/"
#define CRASH_PATH_SUBDIR    "/dump/"
#define LIVE_PATH_SUBDIR     "/livedump/"

#define WAIT_FOR_OPT_TIMEOUT_SEC 60

#define MINICOREDUMPER_TIMEOUT_MS  DEFAULT_COMMAND_TIMEOUT_MS
#define LIVEDUMPER_TIMEOUT_MS      DEFAULT_COMMAND_TIMEOUT_MS
#define CRASH_STACK_TIMEOUT_MS     DEFAULT_COMMAND_TIMEOUT_MS
#define ZIP_TIMEOUT_MS             DEFAULT_COMMAND_TIMEOUT_MS

enum {
	RET_EXCEED = 1,
	NUM_EXCEED,
	USAGE_EXCEED
};

struct file_info {
	bool   isdir;
	size_t size;
	time_t mtime;
	char   *path;
};

/* Configuration variables */
config_t config;
static char* crash_dump_path;
static char* crash_temp_path;

/* pkgmgrinfo filter list function for getting application ID */
static int appinfo_get_appid_func(pkgmgrinfo_appinfo_h handle,
		void *user_data)
{
	char *str = NULL;

	int ret = pkgmgrinfo_appinfo_get_appid(handle, &str);
	if (ret == PMINFO_R_OK && str)
		(*(char **)user_data) = strdup(str);

	return ret;
}

/* get application ID by pkgmgrinfo filter */
static int get_appid(char *exepath, char *appid, int len)
{
	pkgmgrinfo_appinfo_filter_h handle = NULL;
	int count, ret;
	char *aid = NULL;

	ret = pkgmgrinfo_appinfo_filter_create(&handle);
	if (ret != PMINFO_R_OK) {
		ret = -1;
		goto out;
	}

	ret = pkgmgrinfo_appinfo_filter_add_string(handle, PMINFO_APPINFO_PROP_APP_EXEC, exepath);
	if (ret != PMINFO_R_OK) {
		ret = -1;
		goto out_free;
	}

	ret = pkgmgrinfo_appinfo_filter_count(handle, &count);
	if (ret != PMINFO_R_OK) {
		ret = -1;
		goto out_free;
	}

	if (count < 1) {
		ret = -1;
		goto out_free;
	}

	ret = pkgmgrinfo_appinfo_filter_foreach_appinfo(handle, appinfo_get_appid_func, &aid);
	if (ret != PMINFO_R_OK) {
		ret = -1;
		goto out_free;
	}
	if (aid) {
		snprintf(appid, len, "%s", aid);
		ret = 0;
		free(aid);
	}

out_free:
	pkgmgrinfo_appinfo_filter_destroy(handle);
out:
	return ret;
}

/* get package ID by appid */
static int get_pkgid(char *appid, char *pkgid, int len)
{
	pkgmgrinfo_appinfo_h handle = NULL;
	int ret;
	char *pkid = NULL;

	ret = pkgmgrinfo_appinfo_get_appinfo(appid, &handle);
	if (ret != PMINFO_R_OK) {
		ret = -1;
		goto out;
	}
	ret = pkgmgrinfo_appinfo_get_pkgid(handle, &pkid);
	if (ret != PMINFO_R_OK) {
		ret = -1;
		goto out_free;
	}
	snprintf(pkgid, len, "%s", pkid);

out_free:
	pkgmgrinfo_appinfo_destroy_appinfo(handle);
out:
	return ret;
}

static int prepare_paths(struct crash_info* cinfo)
{
	int tmp_len;
	const char *crash_subdir = cinfo->livedump ? LIVE_PATH_SUBDIR : CRASH_PATH_SUBDIR;

	if (cinfo->output_path) {
		free(config.crash_root_path);
		config.crash_root_path = strdup(cinfo->output_path);
		if (!config.crash_root_path) {
			_E("strdup() error: %m");
			return 0;
		}
	}

	tmp_len = strlen(config.crash_root_path) + strlen(crash_subdir);
	crash_dump_path = (char*)malloc(tmp_len + 1);
	if (crash_dump_path == NULL) {
		_E("Couldn't allocate memory for crash_dump_path: %m\n");
		return 0;
	}
	snprintf(crash_dump_path, tmp_len + 1, "%s%s", config.crash_root_path, crash_subdir);

	tmp_len = strlen(config.crash_root_path) + strlen(CRASH_TEMP_SUBDIR);
	crash_temp_path = (char*)malloc(tmp_len + 1);
	if (crash_temp_path == NULL) {
		_E("Couldn't allocate memory for crash_temp_path: %m\n");
		return 0;
	}
	snprintf(crash_temp_path, tmp_len + 1, "%s%s", config.crash_root_path, CRASH_TEMP_SUBDIR);
	return 1;
}

static bool make_dump_dir(void)
{
	const char *dirs[] = {crash_dump_path, crash_temp_path};

	for (size_t i = 0; i < ARRAY_SIZE(dirs); i++) {
		const char *dirname = dirs[i];

		int r = mkdir(dirname, 0775);
		if (r >= 0)
			continue;

		if (errno != EEXIST) {
			_E("Unable to create directory %s: %m", dirname);
			return false;
		}

		struct stat st = {0};
		r = stat(dirname, &st);
		bool isdir = !!(st.st_mode & S_IFDIR);

		if (!r && isdir)
			continue;
		else if (!r && !isdir)
			errno = ENOTDIR;

		_E("Failure while trying to ensure %s exists and it is directory: %m", dirname);
		return false;
	}

	return true;
}

static bool get_cmd_info(struct crash_info *cinfo)
{
	assert(cinfo);

	cinfo->cmd_line = malloc(PATH_MAX);
	if (!cinfo->cmd_line)
		return false;

	if (!read_proc_file(cinfo->pid_info, "cmdline", cinfo->cmd_line, PATH_MAX, NULL))
		goto err;

	char buf[PATH_MAX];
	if (!get_exe_path(cinfo->pid_info, buf, sizeof(buf)))
		goto err;

	cinfo->cmd_path = strdup(buf);

	return !!cinfo->cmd_path;

err:
	free(cinfo->cmd_line);
	cinfo->cmd_line = NULL;
	return false;
}

static int set_prstatus(struct crash_info *cinfo)
{
	int ret;
	char prstatus_name[NAME_MAX+1];

	ret = snprintf(prstatus_name, NAME_MAX, "/%d.prstatus", cinfo->pid_info);
	if (ret < 0) {
		_E("Failed to snprintf for prstatus path");
		goto close_fd;
	}

	cinfo->prstatus_fd = shm_open(prstatus_name, O_RDWR | O_CREAT, 0600);
	if (cinfo->prstatus_fd < 0) {
		_E("shm_open: %m");
		goto close_fd;
	}

	ret = shm_unlink(prstatus_name);
	if (ret < 0) {
		_E("shm_unlink: %m");
		goto close_fd;
	}

	ret = fcntl(cinfo->prstatus_fd, F_GETFD);
	if (ret < 0) {
		_E("fcntl(): %m");
		goto close_fd;
	}

	ret = fcntl(cinfo->prstatus_fd, F_SETFD, ret & ~FD_CLOEXEC);
	if (ret < 0) {
		_E("fcntl(): %m");
		goto close_fd;
	}

	ret = ftruncate(cinfo->prstatus_fd, sizeof(struct elf_prstatus));
	if (ret < 0) {
		_E("ftruncate(): %m");
		goto close_fd;
	}

	return 0;

close_fd:
	if (cinfo->prstatus_fd >= 0)
		close(cinfo->prstatus_fd);
	return -1;
}

bool set_crash_info(struct crash_info *cinfo)
{
	int ret;
	char *temp_dir_ret = NULL;
	char date[80];
	struct tm loc_tm;

	if (cinfo->livedump) {
		if (cinfo->kill)
			cinfo->sig_info = 9;
		else
			cinfo->sig_info = 0;
	}

	if (cinfo->livedump && !file_exists(LIVEDUMPER_BIN_PATH)) {
		fprintf(stderr, "Error: %s doesn't exist - can not perform livedump. Terminating.\n", LIVEDUMPER_BIN_PATH);
		_E("Error: %s doesn't exist - can not perform livedump. Terminating.\n", LIVEDUMPER_BIN_PATH);
		return false;
	}

	if (cinfo->tid_info == -1) {
		if (cinfo->livedump) {
			cinfo->tid_info = cinfo->pid_info;
		} else {
			cinfo->tid_info = find_crash_tid(cinfo->pid_info);
			if (cinfo->tid_info < 0) {
				_I("TID not found");
				cinfo->tid_info = cinfo->pid_info;
			}
		}
	}

	if (!get_cmd_info(cinfo)) {
		_E("Failed to get command info");
		return false;
	}

	if (cinfo->time_info == 0)
		cinfo->time_info = time(NULL);

	localtime_r(&cinfo->time_info, &loc_tm);
	strftime(date, sizeof(date), "%Y%m%d%H%M%S", &loc_tm);

	if (asprintf(&cinfo->temp_dir, "%s/crash.XXXXXX", crash_temp_path) == -1) {
		_E("Failed to asprintf for temp_dir");
		cinfo->temp_dir = NULL;
		return false;
	}

	temp_dir_ret = mkdtemp(cinfo->temp_dir);
	if (!temp_dir_ret || access(temp_dir_ret, F_OK)) {
		_E("Failed to mkdtemp for temp_dir");
		return false;
	}

	if (asprintf(&cinfo->name, "%s_%d_%s", basename(cinfo->cmd_line),
			cinfo->pid_info, date) == -1) {
		_E("Failed to snprintf for name");
		cinfo->name = NULL;
		goto rm_temp;
	}

	if (config.allow_zip)
		ret = asprintf(&cinfo->result_path,
				"%s/%s.zip", crash_dump_path, cinfo->name);
	else
		ret = asprintf(&cinfo->result_path,
				"%s/%s", crash_dump_path, cinfo->name);
	if (ret == -1) {
		_E("Failed to asprintf for result path");
		cinfo->result_path = NULL;
		goto rm_temp;
	}

	if (asprintf(&cinfo->pfx, "%s/%s", cinfo->temp_dir, cinfo->name) == -1) {
		_E("Failed to asprintf for pfx");
		cinfo->pfx = NULL;
		goto rm_temp;
	}
	ret = mkdir(cinfo->pfx, 0775);
	if (ret < 0) {
		_E("Failed to mkdir for %s", cinfo->pfx);
		goto rm_temp;
	}

	if (asprintf(&cinfo->info_path, "%s/%s.info", cinfo->pfx, cinfo->name) == -1) {
		_E("Failed to asprintf for info path");
		cinfo->info_path = NULL;
		goto rm_temp;
	}

	if (asprintf(&cinfo->core_path, "%s/%s.coredump", cinfo->pfx, cinfo->name) == -1) {
		_E("Failed to asprintf for core path");
		cinfo->core_path = NULL;
		goto rm_temp;
	}

	if (asprintf(&cinfo->log_path, "%s/%s.log", cinfo->pfx, cinfo->name) == -1) {
		_E("Failed to asprintf for log path");
		cinfo->log_path = NULL;
		goto rm_temp;
	}

	if (set_prstatus(cinfo) < 0)
		goto rm_temp;

	if (get_appid(cinfo->cmd_line, cinfo->appid, sizeof(cinfo->appid)) < 0) {
		snprintf(cinfo->appid, sizeof(cinfo->appid), "%s", basename(cinfo->cmd_line));
		snprintf(cinfo->pkgid, sizeof(cinfo->pkgid), "%s", basename(cinfo->cmd_line));
	} else {
		if (get_pkgid(cinfo->appid, cinfo->pkgid, sizeof(cinfo->pkgid)) < 0)
			snprintf(cinfo->pkgid, sizeof(cinfo->pkgid), "%s", cinfo->appid);
	}

	return true;

rm_temp:
	remove_dir(cinfo->temp_dir, 1);
	return false;
}

static void launch_crash_popup(struct crash_info *cinfo)
{
	assert(cinfo);

	char *av[] = { CRASH_POPUP_BIN_PATH,
			"--cmdline", cinfo->cmd_line,
			"--cmdpath", cinfo->cmd_path,
			NULL };

	spawn_param_s param = { .fn = spawn_nullstdfds };
	spawn(av, NULL, &param, NULL, NULL);
}

static bool dump_system_state(const struct crash_info *cinfo, pid_t *pid)
{
	char *av[] = {DUMP_SYSTEMSTATE_BIN_PATH, "-d", "-k", "-j", "-p", "-e", "-f",  cinfo->log_path, NULL};
	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	return spawn(av, NULL, &param, pid, NULL);
}

static void save_so_info(const struct crash_info *cinfo)
{
	char maps_path[PATH_MAX];
	char so_info_path[PATH_MAX];
	snprintf(maps_path, sizeof(maps_path), "/proc/%d/maps",
		 cinfo->pid_info);

	snprintf(so_info_path, sizeof(so_info_path),
			 "%s/%s.%s", cinfo->pfx, cinfo->name, "so_info");

	get_and_save_so_info(maps_path, so_info_path);
}

/* remove a file whose name begins with 'beggining_of_name'. This is needed
 * when we don't want to include coredump in report, but we only know the
 * beginning of the coredump file name. */
static int remove_file_in_dir(const char *directory, const char *beginning_of_name)
{
	int ret = -1;
	int errno_unlink = 0;

	DIR *dir = opendir(directory);
	if (dir == NULL)
		goto end;

	int dir_fd = dirfd(dir);
	if (dir_fd == -1) {
		closedir(dir);
		goto end;
	}

	struct dirent* file_info;
	while ((file_info = readdir(dir)) != NULL) {
		if (strstr(file_info->d_name, beginning_of_name) == file_info->d_name) {
			ret = unlinkat(dir_fd, file_info->d_name, 0);
			errno_unlink = errno;
			break;
		}
	}
	closedir(dir);
	errno = errno_unlink; /* for %m */
end:
	return ret;
}

// These macros are used in functions below
#define SNPRINTF_OR_EXIT_W(name, format, member) do { if (snprintf(name##_str, sizeof(name##_str), format, cinfo->member) < 0) goto out; } while (0)
#define SNPRINTF_OR_EXIT(name, format) SNPRINTF_OR_EXIT_W(name, format, name##_info)

static bool extra_script(const struct crash_info *cinfo, pid_t *pid)
{
	if (!config.extra_script)
		return false;

	char pid_str[11];
	SNPRINTF_OR_EXIT(pid, "%d");

	char *av[] = { config.extra_script, cinfo->pfx, pid_str, NULL };
	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	return spawn(av, NULL, &param, pid, NULL);

out:
	return false;
}

static void launch_dbus_notify(struct crash_info *cinfo)
{
	assert(cinfo);

	char pid_str[11], tid_str[11], sig_str[11];
	char *prstatus_fd_str = NULL;
	char tid_comm_str[KERNEL_DEFINED_TASK_COMM_LEN + 1] = { 0, };

	(void)read_proc_file(cinfo->tid_info, "comm", tid_comm_str, sizeof(tid_comm_str), filter_drop_trailing_whitespace);

	if (asprintf(&prstatus_fd_str, "%d", cinfo->prstatus_fd) == -1) {
		_E("Unable to allocate memory: %m");
		return;
	}

	SNPRINTF_OR_EXIT(pid, "%d");
	SNPRINTF_OR_EXIT(tid, "%d");
	SNPRINTF_OR_EXIT(sig, "%d");

	char *av[] = { CRASH_NOTIFY_BIN_PATH,
			"--cmdline", cinfo->cmd_line,
			"--cmdpath", cinfo->cmd_path,
			"--pid", pid_str,
			"--tid", tid_str,
			"--appid", cinfo->appid,
			"--pkgid", cinfo->pkgid,
			"--reportpath", cinfo->result_path,
			"--prstatus_fd", prstatus_fd_str,
			"--signal", sig_str,
			"--tid-comm", tid_comm_str,
			NULL };

	spawn(av, NULL, NULL, NULL, NULL);
out:
	free(prstatus_fd_str);
}

static bool execute_minicoredump(struct crash_info *cinfo, int *exit_code)
{
	char *coredump_name = NULL;
	char *prstatus_fd_str = NULL;
	bool is_ok = false;

	if (asprintf(&coredump_name, "%s.coredump", cinfo->name) == -1
	    || asprintf(&prstatus_fd_str, "%d", cinfo->prstatus_fd) == -1) {
		_E("Unable to allocate memory");
		goto out;
	}

	char pid_str[11], uid_str[11], gid_str[11], sig_str[11], time_str[11];

	SNPRINTF_OR_EXIT(pid, "%d");
	SNPRINTF_OR_EXIT(uid, "%d");
	SNPRINTF_OR_EXIT(gid, "%d");
	SNPRINTF_OR_EXIT(sig, "%d");
	SNPRINTF_OR_EXIT(time, "%ld");

	const char *without_core_str = config.dump_core ? NULL : "-s";

	/* Execute minicoredumper */
	char *args[] = {
			MINICOREDUMPER_BIN_PATH,   // minicoredumper filename path
			pid_str,                   // %p - pid
			uid_str,                   // %u - UID
			gid_str,                   // %g - GID
			sig_str,                   // %s - number of signal
			time_str,                  // %t - time of dump
			"localhost",               // %h - hostname
			basename(cinfo->cmd_line), // %e - exe name (need for result filename)
			MINICOREDUMPER_CONFIG_PATH, // config file
			"-d",
			cinfo->pfx,            // temp dir
			"-o",
			coredump_name,             // coredump filename
			"-P",
			prstatus_fd_str,
			(char *)without_core_str,  // with or without core
			NULL
			};

	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	is_ok = spawn_wait(args, NULL, &param, MINICOREDUMPER_TIMEOUT_MS, exit_code);

	/* Minicoredumper must be executed to dump at least PRSTATUS for
	   other tools, coredump, however, might have been disabled. */
	if (!config.dump_core && file_exists_in_dir(cinfo->pfx, coredump_name)) {
		if (remove_file_in_dir(cinfo->pfx, coredump_name) != 0)
			_E("Saving core disabled - removing coredump %s/%s failed: %m",
			   cinfo->pfx, coredump_name);
		else
			_D("Saving core disabled - removed coredump %s/%s",
			   cinfo->pfx, coredump_name);
	}

out:
	free(coredump_name);
	free(prstatus_fd_str);

	return is_ok;
}

static bool execute_livedumper(const struct crash_info *cinfo, int *exit_code)
{
	char *coredump_path = NULL;
	char *prstatus_fd_str = NULL;
	bool is_ok = false;
	char pid_str[11];

	if (asprintf(&coredump_path, "%s/%s.coredump", cinfo->pfx, cinfo->name) == -1 ||
	    asprintf(&prstatus_fd_str, "%d", cinfo->prstatus_fd) == -1) {
		_E("Unable to allocate memory");
		goto out;
	}

	SNPRINTF_OR_EXIT(pid, "%d");

	/* Execute livedumper */
	char *args[] = {
			LIVEDUMPER_BIN_PATH,    // livedumper filename path
			"-m",                   // minilivecoredump
			"-P", prstatus_fd_str,
			"-f", coredump_path,
			pid_str,                // %p - pid
			NULL
			};

	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	is_ok = spawn_wait(args, NULL, &param, LIVEDUMPER_TIMEOUT_MS, exit_code);
out:
	free(prstatus_fd_str);
	free(coredump_path);
	return is_ok;
}

static bool execute_crash_stack(const struct crash_info *cinfo, int *exit_code)
{
	char pid_str[11], tid_str[11], sig_str[11], prstatus_fd_str[11];
	bool is_ok = false;

	SNPRINTF_OR_EXIT(pid, "%d");
	SNPRINTF_OR_EXIT(tid, "%d");
	SNPRINTF_OR_EXIT(sig, "%d");
	SNPRINTF_OR_EXIT_W(prstatus_fd, "%d", prstatus_fd);

	/* Execute crash-stack */
	char *args[] = { CRASH_STACK_BIN_PATH,
			 "--pid", pid_str,
			 "--tid", tid_str,
			 "--sig", sig_str,
			 "--prstatus_fd", prstatus_fd_str,
			 NULL };

	int fd = open(cinfo->info_path, O_WRONLY | O_CREAT, 0600);
	if (fd < 0) {
		_E("open %s error: %m", cinfo->info_path);
		return false;
	}

	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = fd };
	is_ok = spawn_wait(args, NULL, &param, CRASH_STACK_TIMEOUT_MS, exit_code);
	close(fd);

out:
	return is_ok;
}

#undef SNPRINTF_OR_EXIT
#undef SNPRINTF_OR_EXIT_W

static bool kill_pid(const pid_t pid, int sig, bool wait)
{
	if (kill(pid, sig) == -1) {
		_E("kill sig %d error: %m\n", sig);
		return false;
	}

	if (wait) {
		if (waitpid(pid, NULL, 0) == -1) {
			_E("waitpid error: %m");
			return false;
		}
	}
	return true;
}

static bool process_stop(const pid_t pid)
{
	_D("stop process %d", pid);
	return kill_pid(pid, SIGSTOP, false);
}

static bool process_continue(const pid_t pid)
{
	_D("continue process %d", pid);
	return kill_pid(pid, SIGCONT, false);
}

static bool execute_crash_modules(struct crash_info *cinfo)
{
	int exit_code = 0;
	if (cinfo->livedump) {
		_I("Starting the livedumper");
		if (!process_stop(cinfo->pid_info))
			return false;
		if (!execute_livedumper(cinfo, &exit_code) || exit_code != 0) {
			_E("Failed to run livedumper - can not continue");
			process_continue(cinfo->pid_info);
			return false;
		}
	} else {
		_I("Starting the minicoredumper");
		if (!execute_minicoredump(cinfo, &exit_code) || exit_code != 0) {
			_E("Failed to run minicoredumper - can not continue");
			return false;
		}
	}

	execute_crash_stack(cinfo, NULL);
	if (cinfo->livedump)
		process_continue(cinfo->pid_info);

	return true;
}

static int lock_dumpdir(void)
{
	int fd;

	if ((fd = open(crash_dump_path, O_RDONLY | O_DIRECTORY)) < 0) {
		_E("Failed to open %s: %m", crash_dump_path);
		return -1;
	}

	if (flock(fd, LOCK_EX) < 0) {
		_E("Failed to lock %s for exclusive access: %m", crash_dump_path);
		close(fd);
		return -1;
	}

	return fd;
}

static void unlock_dumpdir(int fd)
{
	if (flock(fd, LOCK_UN) < 0)
		_E("Failed to unlock file descriptor: %m");
	close(fd);
}

static int dump_filter(const struct dirent *de)
{
	if (de->d_name[0] == '.')
		return 0;
	return 1;
}

static int mtime_cmp(const void *_a, const void *_b)
{
	const struct file_info *a = _a;
	const struct file_info *b = _b;

	if (a->mtime < b->mtime)
		return -1;
	if (a->mtime > b->mtime)
		return 1;
	return 0;
}

static int scan_dump(struct file_info **dump_list, off_t *usage)
{
	struct file_info *temp_list;
	struct dirent **scan_list = NULL;
	struct stat st;
	int i, scan_num, dump_num = 0;
	int fd;

	if ((fd = open(crash_dump_path, O_DIRECTORY)) < 0) {
		_E("Failed to open %s: %m", crash_dump_path);
		return -1;
	}

	scan_num = scandir(crash_dump_path, &scan_list, &dump_filter, NULL);
	if (scan_num < 0) {
		close(fd);
		return -1;
	}

	temp_list = (struct file_info *)calloc(scan_num,
			sizeof(struct file_info));
	if (!temp_list) {
		_E("Failed to calloc for dump list");
		goto exit;
	}

	for (i = 0; i < scan_num; i++) {
		if (fstatat(fd, scan_list[i]->d_name, &st, 0) < 0) {
			_E("Failed to fstatat");
			continue;
		}

		if (asprintf(&(temp_list[dump_num].path), "%s/%s",
					crash_dump_path, scan_list[i]->d_name) < 0) {
			_E("Failed to asprintf");
			continue;
		}

		if (scan_list[i]->d_type == DT_DIR) {
			temp_list[dump_num].isdir = 1;
			temp_list[dump_num].size =
				get_directory_usage(temp_list[dump_num].path);
		} else {
			temp_list[dump_num].isdir = 0;
			temp_list[dump_num].size = st.st_size;
		}
		temp_list[dump_num].mtime = st.st_mtime;
		dump_num++;
	}

	if (dump_num <= 0) {
		free(temp_list);
		goto exit;
	}

	if (dump_num != scan_num) {
		struct file_info *const tmp = (struct file_info *)realloc(temp_list,
				dump_num * sizeof(struct file_info));
		if (!tmp) {
			free(temp_list);
			dump_num = -1;
			goto exit;
		}
		temp_list = tmp;
	}

	qsort(temp_list, dump_num, sizeof(struct file_info), mtime_cmp);

	*usage = 0;
	for (i = 0; i < dump_num; i++) {
		_D("[%d] path: %s(%s), size: %zu kb, mtime: %s",
				i,
				temp_list[i].path,
				temp_list[i].isdir ? "DIR" : "FILE",
				temp_list[i].size / 1024,
				ctime(&(temp_list[i].mtime)));
		*usage += temp_list[i].size;
	}
	*dump_list = temp_list;
exit:
	for (i = 0; i < scan_num; i++)
		free(scan_list[i]);
	free(scan_list);
	close(fd);

	return dump_num;
}

static int check_disk_available(const char *path, int check_size)
{
	struct statfs lstatfs;
	int avail_size = 0;

	if (!path)
		return -1;

	if (statfs(path, &lstatfs) < 0)
		return -1;
	avail_size = (int)(lstatfs.f_bavail * (lstatfs.f_bsize / 1024));

	if (check_size > avail_size) {
		_I("avail_size is (%d)", avail_size);
		return -1;
	}

	return 0;
}

static int remove_file_info(struct file_info file)
{
	if (file.isdir)
		return remove_dir(file.path, 1);
	else
		return unlink(file.path);
}

static void clean_dump(void)
{
	struct file_info *dump_list = NULL;
	int i, scan_num, dump_num, remove_flag;
	off_t usage = 0;
	time_t cur_time;

	scan_num = scan_dump(&dump_list, &usage);
	if (scan_num <= 0)
		return;

	dump_num = scan_num;

	cur_time = time(NULL);

	for (i = 0; i < scan_num; i++) {
		remove_flag = 0;

		/* Retention time check */
		if (config.max_retention_sec &&
			dump_list[i].mtime > 0 &&
			dump_list[i].mtime + config.max_retention_sec < cur_time)
			remove_flag = RET_EXCEED;
		/* Check the number of dumps */
		else if (config.max_crash_dump &&
			0 < dump_num && config.max_crash_dump < dump_num)
			remove_flag = NUM_EXCEED;
		/* Check the max system use size */
		else if (config.system_max_use &&
			0 < dump_num && config.system_max_use < usage / 1024)
			remove_flag = USAGE_EXCEED;

		switch (remove_flag) {
		case RET_EXCEED:
			_I("Reached the maximum retention time %d, so remove (%s)",
					config.max_retention_sec, dump_list[i].path);
			break;
		case NUM_EXCEED:
			_I("Reached the maximum number of dump %d/%d, so remove (%s)",
					dump_num, config.max_crash_dump,
					dump_list[i].path);
			break;
		case USAGE_EXCEED:
			_I("Reached the maximum disk usage %" PRId64 "/%d kb, so remove (%s)",
					usage / 1024, config.system_max_use,
					dump_list[i].path);
			break;
		default:
			break;
		}

		if (!remove_flag)
			continue;

		if (remove_file_info(dump_list[i]) < 0) {
			_E("Failed to remove %s", dump_list[i].path);
			continue;
		}
		dump_num--;
		usage -= dump_list[i].size;
	}

	/* Check disk free space to keep */
	if (config.system_keep_free &&
			check_disk_available(config.crash_root_path,
					     config.system_keep_free) < 0) {
		_I("Disk is not available! so set the maximum number of dump to 1");
		config.max_crash_dump = 1;
	}

	for (i = 0; i < dump_num; i++)
		free(dump_list[i].path);
	free(dump_list);
}

static void compress(struct crash_info *cinfo)
{
	int ret, lock_fd;
	char zip_path[PATH_MAX];

	ret = snprintf(zip_path, sizeof(zip_path), "%s/report.zip",
			cinfo->temp_dir);
	if (ret < 0) {
		_E("Failed to snprintf for zip path");
		return;
	}

	char *args[] = {"/bin/zip", "-qyr", zip_path, cinfo->name, NULL};
	spawn_param_s param1 = { .fn = spawn_nullstdfds };
	spawn_param_s param0 = { .fn = spawn_chdir, .u.char_ptr = cinfo->temp_dir, .next = &param1 };
	(void)spawn_wait(args, NULL, &param0, ZIP_TIMEOUT_MS, NULL);

	if ((lock_fd = lock_dumpdir()) < 0)
		return;
	if (!rename(zip_path, cinfo->result_path))
		clean_dump();
	else
		_E("Failed to move %s to %s", zip_path, cinfo->result_path);
	unlock_dumpdir(lock_fd);

	if (remove_dir(cinfo->temp_dir, 1) < 0)
		_E("Failed to delete temp directory");
}

static void move_dump_data(const char *from_path, const struct crash_info *cinfo)
{
	int lock_fd;

	if ((lock_fd = lock_dumpdir()) < 0)
		return;
	if (!rename(from_path, cinfo->result_path))
		clean_dump();
	else
		_E("Failed to move %s to %s",
				from_path, cinfo->result_path);
	unlock_dumpdir(lock_fd);

	if (remove_dir(cinfo->temp_dir, 1) < 0)
		_E("Failed to delete temp directory");
}

static int wait_for_opt(unsigned int timeout)
{
	unsigned int count = 0;

	while (check_disk_available(config.crash_root_path, 0) < 0 && count < timeout) {
		log_kmsg("crash-manager: path %s is not available\n", config.crash_root_path);
		sleep(1);
		count++;
	}
	if (count >= timeout) {
		log_kmsg("crash-manager: timeout (%ds) while waiting for %s."
			"Probably /opt is not mounted.\n", timeout, config.crash_root_path);
		return 0;
	}

	return 1;
}

static void free_crash_info(struct crash_info *cinfo)
{
	free(cinfo->cmd_line);
	free(cinfo->cmd_path);
	free(cinfo->temp_dir);
	free(cinfo->result_path);
	free(cinfo->pfx);
	free(cinfo->info_path);
	free(cinfo->core_path);
	free(cinfo->log_path);
}

void crash_info_init(struct crash_info *cinfo)
{
	cinfo->prstatus_fd = -1;
	cinfo->livedump = false;
	cinfo->kill = false;
	cinfo->print_result_path = false;
	cinfo->tid_info = -1;
	cinfo->time_info = 0;
	cinfo->output_path = NULL;
	cinfo->cmd_line = NULL;
	cinfo->cmd_path = NULL;
	cinfo->temp_dir = NULL;
	cinfo->pfx =  NULL;
	cinfo->result_path = NULL;
	cinfo->info_path = NULL;
	cinfo->core_path = NULL;
	cinfo->log_path = NULL;
}

static bool run(struct crash_info *cinfo)
{
	/* Execute processes in parallel */
	static pid_t dump_state_pid = 0, extra_script_pid = 0;
	int debug_mode = access(DEBUGMODE_PATH, F_OK) == 0;

	if (config.report_type >= REP_TYPE_FULL) {
		/* Exec dump_systemstate */
		if (!dump_system_state(cinfo, &dump_state_pid)) {
			_E("Failed to get system state report");
			return false;
		}

		if (config.extra_script && !extra_script(cinfo, &extra_script_pid))
			_W("Failed to call extra script from config");
	}

	/* Exec crash modules */
	if (!execute_crash_modules(cinfo)) {
		_E("Failed to get basic crash information");
		return false;
	}

	if (config.report_type >= REP_TYPE_FULL) {
		/* Save shared objects info (file names, bulid IDs, rpm package names) */
		save_so_info(cinfo);

		/* Wait misc. pids */
		wait_for_pid(dump_state_pid, NULL);
		if (extra_script_pid > 0)
			wait_for_pid(extra_script_pid, NULL);

		/* Tar compression */
		if (config.allow_zip)
			compress(cinfo);
		else
			move_dump_data(cinfo->pfx, cinfo);
	} else {
		free(cinfo->result_path);
		if (asprintf(&cinfo->result_path, "%s/%s.info",
				crash_dump_path, cinfo->name) == -1) {
			cinfo->result_path = NULL;
			_E("asprintf() error: %m");
			return false;
		}
		move_dump_data(cinfo->info_path, cinfo);
	}

	if (cinfo->print_result_path)
		printf("REPORT_PATH=%s\n", cinfo->result_path);

	if (!cinfo->livedump) {
		/* Release the core pipe as passed by kernel, allowing another
		 * coredump to be handled.
		 *
		 * Due to usage of core_pipe_limit there is limited number of
		 * crash-manager processes that kernel is going to invoke
		 * concurrently.  As the next and last step is a _synchronous_
		 * call to crash-popup we close the descriptor here.
		 *
		 * Note: for VIP processes this will likely cause the system
		 * to reboot without showing popup.
		 */
		close(STDIN_FILENO);

		launch_dbus_notify(cinfo);

		/* launch crash-popup only if the .debugmode file exists */
		if (debug_mode)
			launch_crash_popup(cinfo);
	} else if (cinfo->kill) {
		kill_pid(cinfo->pid_info, SIGKILL, false);
	}

	return true;
}

static bool crash_manager_prepare(struct crash_info *cinfo)
{
	if (!config_init(&config, CRASH_MANAGER_CONFIG_PATH))
		return false;

	if (!prepare_paths(cinfo))
		return false;

	if (!wait_for_opt(WAIT_FOR_OPT_TIMEOUT_SEC))
		return false;

	/* Create crash directories */
	if (!make_dump_dir())
		return false;

	if (!set_crash_info(cinfo))
		return false;

	return true;
}

static void write_dump_reason(const char *reason, const char *base_dir, const char *name)
{
	char *reason_name;

	if (asprintf(&reason_name, "%s.dump_reason", name) == -1) {
		_E("Failed to asprintf for reason_name: %m");
	} else {
		write_to_file(reason, base_dir, reason_name);
		free(reason_name);
	}
}

void crash_manager_free(struct crash_info *cinfo)
{
	if (cinfo->prstatus_fd >= 0)
		close(cinfo->prstatus_fd);
	free(crash_temp_path);
	free(crash_dump_path);
	config_free(&config);

	free_crash_info(cinfo);
}

bool crash_manager_direct(struct crash_info *cinfo)
{
	if (!crash_manager_prepare(cinfo))
		return false;

	return run(cinfo);
}

bool crash_manager_livedump_pid(pid_t pid, const char *dump_reason, char *report_path, size_t report_path_len)
{
	bool result = false;
	struct crash_info cinfo;
	crash_info_init(&cinfo);

	cinfo.livedump = true;
	cinfo.pid_info = pid;

	if (!crash_manager_prepare(&cinfo))
		goto exit;

	write_dump_reason(dump_reason, cinfo.pfx, cinfo.name);

	if (!run(&cinfo))
		goto exit;

	strncpy(report_path, cinfo.result_path, report_path_len);

	result = true;
exit:
	crash_manager_free(&cinfo);
	_I("Exiting with %s", result ? "success" : "fail");
	return result;
}

