/* Utilities for spawning sub-processes
 *
 * Copyright (c) 2018-2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <poll.h>
#include <signal.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <wait.h>

#include "log.h"
#include "util.h"
#include "spawn.h"

/* spawn prepare function(s) - to be called in child process */

int spawn_setstdout(spawn_param_s *param)
{
	assert(param);
	return dup2(param->u.int_val, STDOUT_FILENO) < 0 ? -1 : 0;
}

int spawn_setstderr(spawn_param_s *param)
{
	assert(param);
	return dup2(param->u.int_val, STDERR_FILENO) < 0 ? -1 : 0;
}

int spawn_nullstdfds(spawn_param_s *param)
{
	int fd = open("/dev/null", O_RDWR);
	if (fd < 0)
		return -1;

	int ret = dup2(fd, STDIN_FILENO) < 0 || dup2(fd, STDOUT_FILENO) < 0 || dup2(fd, STDERR_FILENO) < 0 ? -1 : 0;
	if (fd != STDIN_FILENO && fd != STDOUT_FILENO && fd != STDERR_FILENO)
		close(fd);
	return ret;
}

int spawn_chdir(spawn_param_s *param)
{
	assert(param);
	return chdir(param->u.char_ptr);
}

int spawn_umask(spawn_param_s *param)
{
	assert(param);
	(void)umask(param->u.mode_t_val);
	return 0;
}

/* spawn api */

bool wait_for_pid(pid_t pid, int *exit_code)
{
	int status = 0;
	int r = 0;
	bool is_ok;

	if (pid < 0)
		return false;

	do {
		r = waitpid(pid, &status, 0);
		is_ok = r >= 0;
	} while (!is_ok && errno == EINTR);

	if (!is_ok) {
		_E("Error while waiting for child %d status: %m", (int)pid);
		return false;
	}

	/* child has terminated */
	if (WIFEXITED(status))
		r = WEXITSTATUS(status);
	else if (WIFSIGNALED(status))
		r = WTERMSIG(status);
	else if (WIFSTOPPED(status))
		r = WSTOPSIG(status);

	_D("Child with pid %d terminated with exit code %d", (int)pid, r);
	if (exit_code)
		*exit_code = r;

	return true;
}

static int spawn_child(char *const av[], char *const ev[], spawn_param_s *param)
{
	static const int spawn_error = 127;

	while (param) {
		assert(param->fn);

		int r = param->fn(param);
		if (r < 0)
			return spawn_error;

		param = param->next;
	};

	execve(av[0], av, ev);
	return spawn_error;
}

bool spawn(char *const av[], char *const ev[], spawn_param_s *param, pid_t *childpid, int *childfd)
{
	int pipefd[2];
	if (pipe(pipefd) < 0) {
		_E("pipe: %m");
		return false;
	}

	pid_t pid = fork();
	if (pid < 0) {
		_E("fork: %m");
		close(pipefd[0]);
		close(pipefd[1]);
		return false;
	} else if (pid == 0) {
		close(pipefd[0]);
		_exit(spawn_child(av, ev, param));
	}
	close(pipefd[1]);

	char *cmd = concatenate(av);
	char *env = concatenate(ev);
	_D("Spawned child with pid %d to execute command <%s> with environment <%s>", (int)pid, cmd, env);
	free(cmd);
	free(env);

	// parent
	if (childpid)
		*childpid = pid;

	if (childfd) {
		_D("Leaving childfd status descriptor %d open for child status notification", pipefd[0]);
		*childfd = pipefd[0];
	} else
		close(pipefd[0]);

	return true;
}

/* returns true if child terminated */
static bool wait_and_kill(pid_t childpid, int childfd, int timeout_ms)
{
	struct pollfd pfd = { .fd = childfd, .events = POLLIN | POLLERR | POLLHUP };
	int r;

	_D("Beginning to wait %dms for child pid %d", timeout_ms, (int)childpid);
	do {
		r = poll(&pfd, 1, timeout_ms);
	} while (r < 0 && errno == EINTR);

	if (r < 0) {
		_E("Internal error while trying to wait for child pid %d: %m", (int)childpid);
		return false;
	} else if (r == 0) {
		_W("Timeout %dms for child pid %d expired. Killing.", timeout_ms, (int)childpid);
		kill(childpid, SIGKILL);
	} else
		_D("Child pid %d terminated before timeout.", (int)childpid);

	return true;
}

bool spawn_wait(char *const av[], char *const ev[], spawn_param_s *param, int timeout_ms, int *exit_code)
{
	pid_t childpid;
	int childfd;

	if (!spawn(av, ev, param, &childpid, &childfd)) {
		_E("spawn() returned an error - aborting waiting");
		return false;
	}

	if (timeout_ms > 0)
		(void)wait_and_kill(childpid, childfd, timeout_ms);

	bool wait_ok = wait_for_pid(childpid, exit_code);

	close(childfd);

	return wait_ok;
}
